-- Script by W'rkncacnter

---------
-- This section can be modified for specific scenarios to make the in game path look better, if necessary

-- M2/Infinity
EphemeraCollection = 4
EphemeraSequence = 16

-- M1
-- EphemeraCollection = 1
-- EphemeraSequence = 24

-- APX
-- EphemeraCollection = 4
-- EphemeraSequence = 11

-- CollectionsUsed = { 4 } -- If you change the above collection, you may have to uncomment this line and change the number in the list

---------

Triggers = {}
TicksToShow = 10 * 30
UNSET = -500
Hotkey10Pressed = UNSET
Hotkey11Pressed = UNSET
Hotkey12Pressed = UNSET
Hotkey12InputValid = true
MicPressed = UNSET
ObjectiveSetTime = UNSET
ObjectiveType = nil
StartLocation = nil
GoalLocation = nil
CurrentPath = nil
MapToggleState = true
TextMarkers = {} -- Annotations
GameMarkers = {} -- Ephemera

function Triggers.init(restoring)
    if restoring then
        for a in Annotations() do
            if a.text == "<" or a.text == ">" or a.text == "v" or a.text == "^" or a.text == "x" or a.text == "" then
                a.text = ""
                table.insert(TextMarkers, a)
            end
        end
        Game.restore_saved()
    end
    load_level_info()
end

function load_level_info()
    load_level_graph()
    load_item_info()
    load_control_panel_info()
    load_automatic_exit_info()
    load_objectives()
end

function load_level_graph()
    local graph = {}
    local polygon_triggers = {}
    for poly in Polygons() do
        local node = {}
        node.id = poly.index
        node.x = poly.x
        node.y = poly.y
        node.type = poly.type

        if poly.type == "platform" then
            node.plat = Platforms[node.permutation]
        elseif
            poly.type == "light on trigger" or poly.type == "platform on trigger" or poly.type == "light off trigger" or
                poly.type == "platform off trigger"
         then
            table.insert(polygon_triggers, poly)
        end
        node.adjacent = {}

        for adj in poly:adjacent_polygons() do
            table.insert(node.adjacent, adj.index)
        end
        if poly.type == "teleporter" then
            table.insert(node.adjacent, poly.permutation)
        end
        table.insert(graph, node)
    end
    Level._graph = graph
    Level._polygon_triggers = polygon_triggers
end

function load_item_info()
    local items = {}
    for i in Items() do
        if i.type == "key" or i.type == "uplink chip" then
            table.insert(items, i)
        end
    end
    Level._items = items
end

function load_control_panel_info()
    local high_priority_switches = {}
    local low_priority_switches = {}
    local terminals = {}
    for s in Sides() do
        if s.control_panel then
            local cl = s.control_panel.type.class
            if cl == "platform switch" or cl == "tag switch" then
                table.insert(high_priority_switches, {poly = s.polygon, s = s, cp = s.control_panel})
            elseif cl == "light switch" then
                table.insert(low_priority_switches, {poly = s.polygon, s = s, cp = s.control_panel})
            elseif cl == "terminal" then
                table.insert(terminals, {poly = s.polygon, s = s, cp = s.control_panel})
            end
        end
    end
    Level._high_priority_switches = high_priority_switches
    Level._low_priority_switches = low_priority_switches
    Level._terminals = terminals
end

function load_automatic_exit_info()
    local exits = {}
    for poly in Polygons() do
        if poly.type == "automatic exit" then
            table.insert(exits, poly)
        end
    end
    Level._exits = exits
end

function load_objectives()
    local objectives = {}
    table.insert(objectives, {label = "", type = "default"})
    for poly in Polygons() do
        if poly.type == "automatic exit" then
            table.insert(
                objectives,
                {label = string.format("Automatic Exit (%s)", poly.index), poly = poly, type = "exit"}
            )
        elseif poly.type == "must be explored" then
            table.insert(
                objectives,
                {label = string.format("Must Be Explored (%s)", poly.index), poly = poly, type = "explore"}
            )
        elseif poly.type == "light on trigger" then
            table.insert(
                objectives,
                {label = string.format("Light On Trigger (%s)", poly.index), poly = poly, type = "trigger"}
            )
        elseif poly.type == "light off trigger" then
            table.insert(
                objectives,
                {label = string.format("Light Off Trigger (%s)", poly.index), poly = poly, type = "trigger"}
            )
        elseif poly.type == "platform on trigger" then
            table.insert(
                objectives,
                {label = string.format("Platform On Trigger (%s)", poly.index), poly = poly, type = "trigger"}
            )
        elseif poly.type == "platform off trigger" then
            table.insert(
                objectives,
                {label = string.format("Platform Off Trigger (%s)", poly.index), poly = poly, type = "trigger"}
            )
        end
    end

    for i in Items() do
        if i.type == "uplink chip" then
            table.insert(
                objectives,
                {label = string.format("Uplink Chip (%s)", i.polygon.index), poly = i.polygon, type = "uplink"}
            )
        elseif i.type == "key" then
            table.insert(
                objectives,
                {label = string.format("Key (%s)", i.polygon.index), poly = i.polygon, type = "key"}
            )
        end
    end

    for s in Sides() do
        if s.control_panel then
            table.insert(
                objectives,
                {
                    label = string.format("%s (%s)", s.control_panel.type.class.mnemonic, s.polygon.index),
                    poly = s.polygon,
                    side = s,
                    type = "switch" -- Meh
                }
            )
        end
    end

    table.insert(objectives, {label = "Monsters", type = "monster"})
    table.insert(objectives, {label = "Regular Items", type = "item"})

    Level._objectives = objectives
    Level._current_objective = 1
end

function Triggers.idle()
    if reload_items then
        reload_items = false
        load_item_info()
        load_objectives()
    end

    for p in Players() do
        for p in Players() do
            p.overlays[2].text = " "
            p.overlays[2].color = 1
            p.overlays[3].text = " "
        end
        p.monster.polygon._visited = true
        if p.hotkey == 10 and Hotkey10Pressed + TicksToShow < Game.ticks then
            clear_indicators()
            Hotkey11Pressed = UNSET
            Hotkey12Pressed = UNSET
            MicPressed = UNSET
            Hotkey10Pressed = Game.ticks
            StartLocation = p.monster.polygon
            save_map_state()

            -- Allow the user to skip a certain item or monster
            local current_objective = Level._objectives[Level._current_objective]
            if p.action_flags.microphone_button and current_objective.type == "item" then
                if CurrentPath then
                    local goal_poly = Polygons[CurrentPath[#CurrentPath].id]
                    goal_poly._visited = true
                end

                p.action_flags.microphone_button = false
            end

            if p.action_flags.microphone_button and current_objective.type == "monster" then
                if CurrentPath then
                    local goal_poly = Polygons[CurrentPath[#CurrentPath].id]
                    for m in goal_poly:monsters() do
                        m._ignored = true
                    end
                end

                p.action_flags.microphone_button = false
            end
        elseif p.hotkey == 11 and Hotkey11Pressed + TicksToShow < Game.ticks then
            clear_indicators()
            Hotkey10Pressed = UNSET
            Hotkey12Pressed = UNSET
            MicPressed = UNSET
            Hotkey11Pressed = Game.ticks
            StartLocation = p.monster.polygon
            save_map_state()
        elseif p.hotkey == 12 and Hotkey12InputValid then
            Hotkey10Pressed = UNSET
            Hotkey11Pressed = UNSET
            Hotkey12Pressed = UNSET
            MicPressed = UNSET
            clear_indicators()

            if p.action_flags.microphone_button then
                Level._current_objective = 1
                p.action_flags.microphone_button = false
            else
                increment_current_objective()
                local current_objective = Level._objectives[Level._current_objective]
                while current_objective.type == "explore" and current_objective.poly.type ~= "must be explored" do
                    increment_current_objective()
                    current_objective = Level._objectives[Level._current_objective]
                end
                if current_objective.poly then
                    current_objective.poly._visible_on_automap = current_objective.poly.visible_on_automap
                end
                Hotkey12Pressed = Game.ticks
                Hotkey12InputValid = false
                StartLocation = p.monster.polygon
            end
            CurrentPath = nil
            ObjectiveType = nil
        elseif p.action_flags.microphone_button and MicPressed + TicksToShow < Game.ticks and CurrentPath then
            MicPressed = Game.ticks
        end

        -- Latch hotkey 12
        if p.hotkey == 0 and (Game.ticks - Hotkey12Pressed) % 3 == 0 then
            Hotkey12InputValid = true
        end
    end

    if Hotkey10Pressed + TicksToShow > Game.ticks then
        local new_objective = false
        if ObjectiveSetTime ~= Hotkey10Pressed then
            if Level._current_objective == 1 then
                find_objective(StartLocation, standard_valid_node_func)
            else
                find_override_objective(StartLocation, standard_valid_node_func)
            end
            ObjectiveSetTime = Hotkey10Pressed
            new_objective = true
        end
        display_path(new_objective)
    elseif Hotkey11Pressed + TicksToShow > Game.ticks then
        local new_objective = false
        if ObjectiveSetTime ~= Hotkey11Pressed then
            if Level._current_objective == 1 then
                find_objective(StartLocation, less_restrictive_valid_node_func)
            else
                find_override_objective(StartLocation, less_restrictive_valid_node_func)
            end
            ObjectiveSetTime = Hotkey11Pressed
            new_objective = true

        -- if CurrentPath then
        --     print_issues_with_path()
        -- end
        end
        display_path(new_objective)
    elseif MicPressed + TicksToShow > Game.ticks then
        local new_objective = false
        if ObjectiveSetTime ~= MicPressed then
            ObjectiveSetTime = MicPressed
            new_objective = true
        end
        display_path(new_objective)
    end

    if
        Game.ticks == Hotkey10Pressed + TicksToShow + 1 or Game.ticks == Hotkey11Pressed + TicksToShow + 1 or
            Game.ticks == MicPressed + TicksToShow + 1
     then
        clear_indicators()
    end

    for p in Players() do
        if Level._current_objective ~= 1 then
            p.overlays[4].color = "yellow"
            local objective = Level._objectives[Level._current_objective]
            if
                objective.poly and objective.type == "switch" and
                    (not objective.side.control_panel or
                        (objective.side.control_panel and objective.side.control_panel._visited))
             then
                p.overlays[4].text = string.format("Override objective: %s (previously used)", objective.label)
            elseif objective.poly and objective.type == "trigger" and objective.poly._visited then
                p.overlays[4].text = string.format("Override objective: %s (previously visited)", objective.label)
            else
                p.overlays[4].text = string.format("Override objective: %s", objective.label)
            end
        else
            p.overlays[4]:clear()
        end
    end

    draw_override_objective()
end

function increment_current_objective()
    Level._current_objective = Level._current_objective + 1
    if Level._current_objective > #Level._objectives then
        Level._current_objective = 1
    end
end

function display_path(new_objective)
    if CurrentPath and new_objective then
        local num_annotations_to_add = #CurrentPath - #TextMarkers
        for i = 1, num_annotations_to_add do
            table.insert(TextMarkers, Annotations.new(Polygons[0], ""))
        end
        for i = 1, #CurrentPath do
            local poly = Polygons[CurrentPath[i].id]
            local a = TextMarkers[i]
            a.text = get_directional_annotation(poly, CurrentPath[i + 1])
            a.polygon = poly
            a.x = poly.x
            a.y = poly.y
        end
        draw_game_path()
    end

    for p in Players() do
        if ObjectiveType and CurrentPath then
            p.overlays[0].text = string.format("Tracking %s", ObjectiveType)
        else
            p.overlays[0].text = "No path found"
        end
    end
    draw_map_path()
end

function draw_map_path()
    if CurrentPath then
        if Game.ticks % 15 == 0 then
            MapToggleState = not MapToggleState
        end
        for i = 1, #CurrentPath do
            local poly = Polygons[CurrentPath[i].id]

            if not MapToggleState and Game.ticks % 15 ~= 0 and poly.visible_on_automap then
                -- We told it not to draw this polygon, but it's still being drawn.
                -- Therefore, you must currently be looking at it. Mark it as seen
                poly._visible_on_automap = true
            end

            poly.visible_on_automap = MapToggleState
        end
    end
end

function draw_override_objective()
    if not CurrentPath and Level._current_objective ~= 1 and Level._objectives[Level._current_objective].poly then
        if Game.ticks % 15 == 0 then
            MapToggleState = not MapToggleState
        end
        local poly = Level._objectives[Level._current_objective].poly
        poly.visible_on_automap = MapToggleState
        if #TextMarkers == 0 then
            table.insert(TextMarkers, Annotations.new(poly, "x"))
        else
            TextMarkers[1].polygon = poly
            TextMarkers[1].x = poly.x
            TextMarkers[1].y = poly.y
            TextMarkers[1].text = "x"
        end
    end
end

function draw_game_path()
    if CurrentPath then
        for i = 1, #CurrentPath do
            local poly = Polygons[CurrentPath[i].id]
            local height = .4
            if poly.media and poly.media.height > poly.floor.height and poly.media.height - poly.floor.height < .6 then
                height = (poly.media.height - poly.floor.height) + .1
            end
            if poly.floor.height + height > poly.ceiling.height then
                height = (poly.ceiling.height - poly.floor.height) / 2
            end
            local e = Ephemera.new(poly.x, poly.y, poly.z + height, poly, EphemeraCollection, EphemeraSequence, 0)
            if e then
                e.tiny = true
                table.insert(GameMarkers, e)
            end
        end
    end
end

function get_directional_annotation(poly1, poly2)
    if not poly2 then
        return "x"
    end

    local diff_x = poly1.x - poly2.x
    local diff_y = poly1.y - poly2.y

    if math.abs(diff_x) > math.abs(diff_y) then
        if diff_x < 0 then
            return ">"
        else
            return "<"
        end
    else
        if diff_y < 0 then
            return "v"
        else
            return "^"
        end
    end
end

function Triggers.got_item(type, player)
    if type == "key" or type == "uplink chip" then
        reload_items = true
    end
end

function Triggers.tag_switch(tag, player, side)
    if side.control_panel then
        side.control_panel._visited = true
    end
end

function Triggers.light_switch(light, player, side)
    if side.control_panel then
        side.control_panel._visited = true
    end
end

function Triggers.platform_switch(polygon, player, side)
    if side.control_panel then
        side.control_panel._visited = true
    end
end

function Triggers.projectile_switch(projectile, side)
    if side.control_panel then
        side.control_panel._visited = true
    end
end

function Triggers.terminal_enter(terminal, player)
    if terminal then -- This doesn't seem to get passed in in M1?
        for i = 1, #Level._terminals do
            local term = Level._terminals[i]
            if term.cp.permutation == terminal.index then
                term.cp._visited = true
            end
            local completion_state = Level.calculate_completion_state()
            if
                term.cp.permutation == terminal.index and
                    (completion_state == "finished" or completion_state == "failed")
             then
                term.cp._confirmed_not_goal = true
            end
        end
    end
end

function save_map_state()
    for poly in Polygons() do
        if
            not (Level._current_objective ~= 1 and Level._objectives[Level._current_objective].poly and
                Level._objectives[Level._current_objective].poly == poly)
         then
            poly._visible_on_automap = poly.visible_on_automap
        end
    end
end

function clear_indicators()
    restore_map_state()
    for p in Players() do
        p.overlays[0]:clear()
    end
end

function restore_map_state()
    if CurrentPath then
        for i = 1, #CurrentPath do
            local poly = Polygons[CurrentPath[i].id]
            poly.visible_on_automap = poly._visible_on_automap
        end
    elseif Level._current_objective and Level._objectives[Level._current_objective].poly then
        Level._objectives[Level._current_objective].poly.visible_on_automap =
            Level._objectives[Level._current_objective].poly._visible_on_automap
    end
    for i = 1, #TextMarkers do
        TextMarkers[i].text = ""
    end

    for i = 1, #GameMarkers do
        local e = GameMarkers[i]
        if e and e.valid then
            e:delete()
        end
    end
    GameMarkers = {}

    for p in Polygons() do
        p._previous_poly_in_path = nil
    end
end

function calc_path(paths, start, goal, type, valid_node_func)
    AttemptedObjectiveType = type
    GoalLocation = Polygons[goal - 1]
    local path, reason =
        get_shortest_path(Level._graph[start], Level._graph[goal], Level._graph, false, valid_node_func)
    if path then
        ObjectiveType = type
        ObjectiveSetTime = Game.ticks
        table.insert(paths, path)
    end
end

function find_objective(start_poly, valid_node_func)
    CurrentPath = nil
    ObjectiveType = nil
    --[[
        Priority order:

        1. Auto exit polygons (if the level is complete)
        2. Unvisited terminals (if the level is complete)
        3. Other terminals (if the level is complete)
        4. Must be explored polygons (if exploration)
        5. Uplink chips / key cards
        6. Unflipped platform/tag switches (above damaging liquid)
        7. Unvisited polygon triggers (above damaging liquid)
        8. Unflipped light switches (above damaging liquid)
        9. Alive monsters (if extermination)
        10. Unvisited terminals
        11. Alive monsters
        12. Unflipped platform/tag switches (below damaging liquid)
        13. Unvisited polygon triggers (below damaging liquid)
        14. Unflipped light switches (below damaging liquid)
    ]]
    local completion_state = Level.calculate_completion_state()

    if completion_state == "finished" or completion_state == "failed" then
        -- Priority 1: Auto exit polygons (if the level is complete)
        if #Level._exits > 0 then
            local paths = {}
            for i = 1, #Level._exits do
                local exit = Level._exits[i]
                local start = start_poly.index + 1
                local goal = exit.index + 1
                calc_path(paths, start, goal, "automatic exit", valid_node_func)
            end
            CurrentPath = best_path(paths)
            if CurrentPath then
                return
            end
        end

        -- Priority 2: Unvisited terminals (if the level is complete)
        set_path_to_unvisited_terminals(start_poly, valid_node_func)
        if CurrentPath then
            return
        end

        -- Priority 3: Other terminals (if the level is complete)
        if #Level._terminals > 0 then
            local paths = {}
            for i = 1, #Level._terminals do
                if not Level._terminals[i].cp._confirmed_not_goal and is_functional(Level._terminals[i].s) then
                    local term_poly = Level._terminals[i].poly
                    local start = start_poly.index + 1
                    local goal = term_poly.index + 1
                    calc_path(paths, start, goal, "terminal", valid_node_func)
                end
            end
            CurrentPath = best_path(paths)
            if CurrentPath then
                return
            end
        end
    end

    -- Priority 4: Must be explored polygons (if exploration)
    if Level.exploration then
        local paths = {}
        for p in Polygons() do
            if p.type == "must be explored" then
                local start = start_poly.index + 1
                local goal = p.index + 1
                calc_path(paths, start, goal, "must be explored", valid_node_func)
            end
        end
        CurrentPath = best_path(paths)
        if CurrentPath then
            return
        end
    end

    -- Priority 5: Uplink chips / key cards
    if #Level._items > 0 then
        local paths = {}
        for i = 1, #Level._items do
            local item = Level._items[i]
            local start = start_poly.index + 1
            local goal = item.polygon.index + 1
            calc_path(paths, start, goal, "item", valid_node_func)
        end
        CurrentPath = best_path(paths)
        if CurrentPath then
            return
        end
    end

    -- Priority 6: Unflipped platform/tag switches (above damaging liquid)
    if #Level._high_priority_switches > 0 then
        set_path_to_unflipped_high_priority_switches(start_poly, false, valid_node_func)
        if CurrentPath then
            return
        end
    end

    -- Priority 7: Unvisited polygon triggers (above damaging liquid)
    if #Level._polygon_triggers > 0 then
        set_path_to_platform_triggers(start_poly, false, valid_node_func)
        if CurrentPath then
            return
        end
    end

    -- Priority 8: Unflipped light switches (above damaging liquid)
    if #Level._low_priority_switches > 0 then
        set_path_to_unflipped_low_priority_switches(start_poly, false, valid_node_func)
        if CurrentPath then
            return
        end
    end

    -- Priority 9: Alive monsters (if extermination)
    if Level.extermination then
        set_path_to_alive_monsters(start_poly, valid_node_func)
        if CurrentPath then
            return
        end
    end

    -- Priority 10: Unvisited terminals
    set_path_to_unvisited_terminals(start_poly, valid_node_func)
    if CurrentPath then
        return
    end

    -- Priority 11: Alive monsters
    set_path_to_alive_monsters(start_poly, valid_node_func)
    if CurrentPath then
        return
    end

    -- Priority 12: Unflipped platform/tag switches (below damaging liquid)
    if #Level._high_priority_switches > 0 then
        set_path_to_unflipped_high_priority_switches(start_poly, true, valid_node_func)
        if CurrentPath then
            return
        end
    end

    -- Priority 13: Unvisited polygon triggers (below damaging liquid)
    if #Level._polygon_triggers > 0 then
        set_path_to_platform_triggers(start_poly, true, valid_node_func)
        if CurrentPath then
            return
        end
    end

    -- Priority 14: Unflipped light switches (below damaging liquid)
    if #Level._low_priority_switches > 0 then
        set_path_to_unflipped_low_priority_switches(start_poly, true, valid_node_func)
        if CurrentPath then
            return
        end
    end
end

function find_override_objective(start_poly, valid_node_func)
    if start_poly then
        local objective = Level._objectives[Level._current_objective]
        local paths = {}
        if objective.type == "monster" then
            set_path_to_alive_monsters(start_poly, valid_node_func)
            if not CurrentPath then
                ObjectiveType = nil
            end
        elseif objective.type == "item" then
            set_path_to_regular_items(start_poly, valid_node_func)
            if not CurrentPath then
                ObjectiveType = nil
            end
        else
            local start = start_poly.index + 1
            local goal = objective.poly.index + 1
            calc_path(paths, start, goal, objective.type, valid_node_func)
            CurrentPath = best_path(paths)
        end
    end
end

function is_in_damaging_liquid(poly)
    if
        poly.media and (poly.media.type == "lava" or poly.media.type == "goo") and
            poly.media.height >= poly.ceiling.height
     then
        return true
    end
    return false
end

function is_functional(side)
    local cp = side.control_panel
    if not cp then
        return false
    end
    if cp.light_dependent and side.primary.light.intensity <= .75 then
        return false
    end

    return true
end

function set_path_to_unflipped_high_priority_switches(start_poly, in_damaging_liquid, valid_node_func)
    local paths = {}
    for i = 1, #Level._high_priority_switches do
        local switch = Level._high_priority_switches[i]
        -- Control panels can get deleted when destroyed, so we have to make sure the cp reference is still good
        if
            switch.s.control_panel == switch.cp and not switch.cp._visited and is_functional(switch.s) and
                is_in_damaging_liquid(switch.poly) == in_damaging_liquid
         then
            local start = start_poly.index + 1
            local goal = switch.poly.index + 1
            calc_path(paths, start, goal, "switch", valid_node_func)
        end
    end
    CurrentPath = best_path(paths)
end

function set_path_to_unflipped_low_priority_switches(start_poly, in_damaging_liquid, valid_node_func)
    local paths = {}
    for i = 1, #Level._low_priority_switches do
        local switch = Level._low_priority_switches[i]
        -- Control panels can get deleted when destroyed, so we have to make sure the cp reference is still good
        if
            switch.s.control_panel == switch.cp and not switch.cp._visited and is_functional(switch.s) and
                is_in_damaging_liquid(switch.poly) == in_damaging_liquid
         then
            local start = start_poly.index + 1
            local goal = switch.poly.index + 1
            calc_path(paths, start, goal, "switch", valid_node_func)
        end
    end
    CurrentPath = best_path(paths)
end

function set_path_to_platform_triggers(start_poly, in_damaging_liquid, valid_node_func)
    local paths = {}
    for i = 1, #Level._polygon_triggers do
        local poly = Level._polygon_triggers[i]
        if not poly._visited and is_in_damaging_liquid(poly) == in_damaging_liquid then
            local start = start_poly.index + 1
            local goal = poly.index + 1
            calc_path(paths, start, goal, "polygon trigger", valid_node_func)
        end
    end
    CurrentPath = best_path(paths)
    if CurrentPath then
        return
    end
end

function set_path_to_alive_monsters(start_poly, valid_node_func)
    local start = start_poly.index + 1
    local paths = {}
    for m in Monsters() do
        if not m.player and not m._ignored and m.visible and is_enemy(m.type) then
            local goal = m.polygon.index + 1
            calc_path(paths, start, goal, "monster", valid_node_func)

            -- We don't need to calc paths to every monster. Get a small handful and find the closest one out of those
            if #paths >= 5 then
                break
            end
        end
    end

    CurrentPath = best_path(paths)
    if CurrentPath then
        return
    end

    -- If no visible monsters found, try invisible
    for m in Monsters() do
        if not m.player and not m._ignored and not m.visible and is_enemy(m.type) then
            local goal = m.polygon.index + 1
            calc_path(paths, start, goal, "monster", valid_node_func)

            -- We don't need to calc paths to every monster. Get a small handful and find the closest one out of those
            if #paths >= 5 then
                break
            end
        end
    end

    CurrentPath = best_path(paths)
    if CurrentPath then
        return
    end

    -- If nothing has been found yet, try non enemies
    for m in Monsters() do
        if not m.player and not m._ignored and not is_enemy(m.type) then
            local goal = m.polygon.index + 1
            calc_path(paths, start, goal, "monster", valid_node_func)

            -- We don't need to calc paths to every monster. Get a small handful and find the closest one out of those
            if #paths >= 5 then
                break
            end
        end
    end

    CurrentPath = best_path(paths)
    if CurrentPath then
        Players.print("Found no enemies, tracking neutral or friendly monster")
    end
end

function set_path_to_regular_items(start_poly, valid_node_func)
    local start = start_poly.index + 1
    local paths = {}
    for i in Items() do
        if i.visible and not i.polygon._visited then
            local goal = i.polygon.index + 1
            calc_path(paths, start, goal, "item", valid_node_func)

            -- We don't need to calc paths to every item. Get a small handful and find the closest one out of those
            if #paths >= 5 then
                break
            end
        end
    end

    CurrentPath = best_path(paths)

    -- If no visible items found, try invisible
    if not CurrentPath then
        for i in Items() do
            if not i.visible and not i.polygon._visited then
                local goal = i.polygon.index + 1
                calc_path(paths, start, goal, "item", valid_node_func)

                -- We don't need to calc paths to every item. Get a small handful and find the closest one out of those
                if #paths >= 5 then
                    break
                end
            end
        end
    end

    CurrentPath = best_path(paths)
end

function set_path_to_unvisited_terminals(start_poly, valid_node_func)
    local unvisited_terminals = {}
    for i = 1, #Level._terminals do
        local term = Level._terminals[i]
        if not term.cp._visited and is_functional(term.s) then
            table.insert(unvisited_terminals, term)
        end
    end
    local paths = {}
    for i = 1, #unvisited_terminals do
        local term_poly = unvisited_terminals[i].poly
        local start = start_poly.index + 1
        local goal = term_poly.index + 1
        calc_path(paths, start, goal, "terminal", valid_node_func)
    end
    CurrentPath = best_path(paths)
end

-- Just doing shortest path based on straight line distance form start to finish locations
function best_path(paths)
    if #paths == 1 then
        return paths[1]
    end

    local shortest_dist = 9999
    local shortest_path = nil
    for i = 1, #paths do
        local path = paths[i]
        local dist = get_distance(path[1], path[#path])
        if dist < shortest_dist then
            shortest_path = path
            shortest_dist = dist
        end
    end

    return shortest_path
end

function is_enemy(type)
    return type.enemies["player"]
end

function get_distance(p1, p2)
    return math.sqrt((p1.x - p2.x) ^ 2 + (p1.y - p2.y) ^ 2)
end

function standard_valid_node_func(node, neighbor)
    if table_contains(node.adjacent, neighbor.id) then
        local poly = Polygons[node.id]
        local adj_poly = Polygons[neighbor.id]
        return is_valid_path(poly, adj_poly)
    end
    return false
end

function is_valid_path(poly, adj_poly, skip_gap_check)
    adj_poly._previous_poly_in_path = nil

    if poly.type == "teleporter" then
        return true, ""
    end

    if
        (AttemptedObjectiveType == "terminal" or AttemptedObjectiveType == "switch") and
            adj_poly.index == GoalLocation.index
     then
        return true, ""
    end

    if adj_poly.type ~= "platform" and poly.type ~= "platform" and adj_poly.floor.height > poly.floor.height + (1 / 3) then
        -- Floor height is too high, but we need to check for media
        if
            not (poly.media and poly.media.height >= adj_poly.floor.height - .25 and
                poly.media.height - poly.floor.height > .833)
         then -- These numbers are just guesses for when you can jump out of media
            -- Check to see if you can jump here from a previous polygon
            if not poly._previous_poly_in_path or not can_jump(poly._previous_poly_in_path, adj_poly) then
                return false, string.format("floor height too high %s, %s", poly.index, adj_poly.index)
            end
        end
    elseif
        adj_poly.type ~= "platform" and poly.type ~= "platform" and
            adj_poly.ceiling.height < poly.floor.height + MonsterTypes["player"].height
     then
        return false, string.format("ceiling height too low %s, %s", poly.index, adj_poly.index)
    elseif
        adj_poly.type ~= "platform" and adj_poly.ceiling.height - adj_poly.floor.height < MonsterTypes["player"].height
     then
        return false, string.format("player can't fit within polygon %s, %s", poly.index, adj_poly.index)
    elseif not skip_gap_check and is_gap_too_small(poly, adj_poly) then
        return false, string.format("gap too small to fit through %s, %s", poly.index, adj_poly.index)
    elseif all_lines_are_solid(poly, adj_poly) then
        return false, string.format("all lines are solid %s, %s", poly.index, adj_poly.index)
    elseif poly.type == "platform" then
        local plat = Platforms[poly.permutation]
        if not plat.active and not plat.player_controllable then
            if adj_poly.type ~= "platform" and adj_poly.floor.height > plat.floor_height + (1 / 3) then
                -- Floor height is too high, but we need to check for media
                if
                    not (poly.media and poly.media.height >= adj_poly.floor.height - .25 and
                        poly.media.height - poly.floor.height > .833)
                 then -- These numbers are just guesses for when you can jump out of media
                    return false, string.format("floor height too high %s, %s", poly.index, adj_poly.index)
                end
            elseif
                adj_poly.type ~= "platform" and
                    adj_poly.ceiling.height < poly.floor.height + MonsterTypes["player"].height
             then
                return false, string.format("ceiling height too low %s, %s", poly.index, adj_poly.index)
            end
        end
    elseif adj_poly.type == "platform" then
        local plat = Platforms[adj_poly.permutation]
        if plat.minimum_floor_height > poly.floor.height + (1 / 3) then
            return false, string.format("platform floor too high %s, %s", poly.index, adj_poly.index)
        elseif plat.maximum_ceiling_height < poly.floor.height + MonsterTypes["player"].height then
            return false, string.format("platform ceiling too low %s, %s", poly.index, adj_poly.index)
        end
        if not plat.active then
            if not plat.player_controllable then
                if plat.floor_height > poly.floor.height + (1 / 3) then
                    -- BUG: There could be a switch nearby that activates the platform, in which case, it should be a valid part of the path
                    return false, string.format("platform floor too high %s, %s", poly.index, adj_poly.index)
                elseif plat.ceiling_height < poly.floor.height + MonsterTypes["player"].height then
                    -- BUG: There could be a switch nearby that activates the platform, in which case, it should be a valid part of the path
                    return false, string.format("platform ceiling too low %s, %s", poly.index, adj_poly.index)
                end
            elseif not plat.door and plat.floor_height > poly.floor.height + (1 / 3) then
                return false, string.format("platform is currently raised too high %s, %s", poly.index, adj_poly.index)
            end
        end
    end

    if
        poly._previous_poly_in_path and poly._previous_poly_in_path.floor.z > poly.floor.z and
            can_jump(poly._previous_poly_in_path, poly, true) and
            not can_jump(poly._previous_poly_in_path, poly, false, true)
     then
        adj_poly._previous_poly_in_path = poly._previous_poly_in_path
    else
        adj_poly._previous_poly_in_path = poly
    end
    return true, ""
end

function less_restrictive_valid_node_func(node, neighbor)
    if table_contains(node.adjacent, neighbor.id) then
        local poly = Polygons[node.id]
        local adj_poly = Polygons[neighbor.id]
        if poly.type == "teleporter" then
            return true
        end

        if all_lines_are_solid(poly, adj_poly) then
            return false, string.format("all lines are solid %s, %s", poly.index, adj_poly.index)
        elseif
            not (AttemptedObjectiveType == "terminal" or AttemptedObjectiveType == "switch") and
                adj_poly.type ~= "platform" and
                poly.type ~= "platform" and
                adj_poly.ceiling.height < poly.floor.height + MonsterTypes["player"].height
         then
            return false, string.format("ceiling height too low %s, %s", poly.index, adj_poly.index)
        elseif
            not (AttemptedObjectiveType == "terminal" or AttemptedObjectiveType == "switch") and
                adj_poly.type ~= "platform" and
                adj_poly.ceiling.height - adj_poly.floor.height < MonsterTypes["player"].height
         then
            return false, string.format("player can't fit within polygon %s, %s", poly.index, adj_poly.index)
        end

        return true
    end
    return false
end

function can_jump(start_poly, end_poly, longest, use_ceiling_height)
    local dist
    if longest then
        dist = get_longest_distance(start_poly, end_poly)
    else
        dist = get_shortest_distance(start_poly, end_poly)
    end

    local height_difference = start_poly.floor.z - end_poly.floor.z
    if use_ceiling_height then
        height_difference = start_poly.floor.z - end_poly.ceiling.z
    end
    local gravity = 0.0024

    -- Formula from Solra and future_you5
    local max_gap_distance = 0.1468 * (math.sqrt(2 * (height_difference + 0.333) / gravity)) + 0.5
    return dist <= max_gap_distance
end

function get_longest_distance(start_poly, end_poly)
    longest = -9999
    for e1 in start_poly:endpoints() do
        for e2 in end_poly:endpoints() do
            local dist = get_distance(e1, e2)
            if dist > longest then
                longest = dist
            end
        end
    end
    return longest
end

function get_shortest_distance(start_poly, end_poly)
    shortest = 9999
    for e1 in start_poly:endpoints() do
        for e2 in end_poly:endpoints() do
            local dist = get_distance(e1, e2)
            if dist < shortest then
                shortest = dist
            end
        end
    end
    return shortest
end

function print_issues_with_path()
    for i = 1, #CurrentPath - 1 do
        local poly = Polygons[CurrentPath[i].id]
        local adj_poly = Polygons[CurrentPath[i + 1].id]
        is_valid, reason = is_valid_path(poly, adj_poly)
        if not is_valid then
            Players.print(reason)
        end
    end
end

function all_lines_are_solid(p1, p2)
    if p1.type == "platform" or p2.type == "platform" then
        return false -- Something weird with platforms and solid lines
    end
    for line in p1:lines() do
        if (line.cw_polygon == p1 or line.cw_polygon == p2) and (line.ccw_polygon == p1 or line.ccw_polygon == p2) then
            if not line.solid then
                return false
            end
        end
    end
    return true
end

function is_gap_too_small(p1, p2)
    local line_is_too_short = false
    local line = nil

    -- Find the line between the two polygons and see if it's too short
    for l in p1:lines() do
        if (l.cw_polygon == p1 or l.cw_polygon == p2) and (l.ccw_polygon == p1 or l.ccw_polygon == p2) then
            if l.length < .5 then
                line_is_too_short = true
            end
            line = l
            break
        end
    end

    if not line_is_too_short then
        return false
    end

    -- If the line is too short, find adjacent polygons that share the same endpoints as the line
    local polys_to_check = {}
    for poly in p1:adjacent_polygons() do
        if not poly == p2 then
            for endpoint in poly:endpoints() do
                if endpoint == line.endpoints[0] or endpoint == line.endpoints[1] then
                    table.insert(polys_to_check, poly)
                end
            end
        end
    end

    -- Check to see if you can enter the adjacent polygons, and that they connect to p2
    for i = 1, #polys_to_check - 1 do
        local poly = polys_to_check[i]
        for adj_poly in poly:adjacent_polygons() do
            if adj_poly == p2 then
                local route_from_p1_to_poly = is_valid_path(p1, poly, true)
                local route_from_poly_to_p2 = is_valid_path(poly, p2, true)
                if route_from_p1_to_poly and route_from_poly_to_p2 then
                    return false
                end
            end
        end
    end

    return true
end

function table_contains(table, element)
    if not table then
        return false
    end

    for _, value in ipairs(table) do
        if value == element then
            return true
        end
    end
    return false
end

--- a* stuff

-- ======================================================================
-- Copyright (c) 2012 RapidFire Studio Limited
-- All Rights Reserved.
-- http://www.rapidfirestudio.com

-- Permission is hereby granted, free of charge, to any person obtaining
-- a copy of this software and associated documentation files (the
-- "Software"), to deal in the Software without restriction, including
-- without limitation the rights to use, copy, modify, merge, publish,
-- distribute, sublicense, and/or sell copies of the Software, and to
-- permit persons to whom the Software is furnished to do so, subject to
-- the following conditions:

-- The above copyright notice and this permission notice shall be
-- included in all copies or substantial portions of the Software.

-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
-- EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
-- MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
-- IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
-- CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
-- TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
-- SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
-- ======================================================================

-- This code came from https://github.com/lattejed/a-star-lua

----------------------------------------------------------------
-- local variables
----------------------------------------------------------------

local INF = 1 / 0
local cachedPaths = nil

----------------------------------------------------------------
-- local functions
----------------------------------------------------------------

function dist(x1, y1, x2, y2)
    return math.sqrt(math.pow(x2 - x1, 2) + math.pow(y2 - y1, 2))
end

function dist_between(nodeA, nodeB)
    return dist(nodeA.x, nodeA.y, nodeB.x, nodeB.y)
end

function heuristic_cost_estimate(nodeA, nodeB)
    return dist(nodeA.x, nodeA.y, nodeB.x, nodeB.y)
end

function is_valid_node(node, neighbor)
    return true
end

function lowest_f_score(set, f_score)
    local lowest, bestNode = INF, nil
    for _, node in ipairs(set) do
        local score = f_score[node]
        if score < lowest then
            lowest, bestNode = score, node
        end
    end
    return bestNode
end

function neighbor_nodes(theNode, nodes)
    local neighbors = {}
    for _, node in ipairs(nodes) do
        if theNode ~= node and is_valid_node(theNode, node) then
            table.insert(neighbors, node)
        end
    end
    return neighbors
end

function not_in(set, theNode)
    for _, node in ipairs(set) do
        if node == theNode then
            return false
        end
    end
    return true
end

function remove_node(set, theNode)
    for i, node in ipairs(set) do
        if node == theNode then
            set[i] = set[#set]
            set[#set] = nil
            break
        end
    end
end

function unwind_path(flat_path, map, current_node)
    if map[current_node] then
        table.insert(flat_path, 1, map[current_node])
        return unwind_path(flat_path, map, map[current_node])
    else
        return flat_path
    end
end

----------------------------------------------------------------
-- pathfinding functions
----------------------------------------------------------------

function a_star(start, goal, nodes, valid_node_func)
    local closedset = {}
    local openset = {start}
    local came_from = {}

    if valid_node_func then
        is_valid_node = valid_node_func
    end

    local g_score, f_score = {}, {}
    g_score[start] = 0
    f_score[start] = g_score[start] + heuristic_cost_estimate(start, goal)

    while #openset > 0 do
        local current = lowest_f_score(openset, f_score)
        if current == goal then
            local path = unwind_path({}, came_from, goal)
            table.insert(path, goal)
            return path
        end

        remove_node(openset, current)
        table.insert(closedset, current)

        local neighbors = neighbor_nodes(current, nodes)
        for _, neighbor in ipairs(neighbors) do
            if not_in(closedset, neighbor) then
                local tentative_g_score = g_score[current] + dist_between(current, neighbor)

                if not_in(openset, neighbor) or tentative_g_score < g_score[neighbor] then
                    came_from[neighbor] = current
                    g_score[neighbor] = tentative_g_score
                    f_score[neighbor] = g_score[neighbor] + heuristic_cost_estimate(neighbor, goal)
                    if not_in(openset, neighbor) then
                        table.insert(openset, neighbor)
                    end
                end
            end
        end
    end
    return nil -- no valid path
end

----------------------------------------------------------------
-- exposed functions
----------------------------------------------------------------

function clear_cached_paths()
    cachedPaths = nil
end

function distance(x1, y1, x2, y2)
    return dist(x1, y1, x2, y2)
end

function get_shortest_path(start, goal, nodes, ignore_cache, valid_node_func)
    if not cachedPaths then
        cachedPaths = {}
    end
    if not cachedPaths[start] then
        cachedPaths[start] = {}
    elseif cachedPaths[start][goal] and not ignore_cache then
        return cachedPaths[start][goal]
    end

    return a_star(start, goal, nodes, valid_node_func)
end

function cheat()
    Players[0].items["assault rifle"] = 1
    Players[0].items["assault rifle ammo"] = 15
    Players[0].items["assault rifle grenades"] = 8
    Players[0].life = 450
end
